//	The beginnings of the C/C++ Groups & Graphs:
//	Method definitions for a generic character buffer

#include "CharBuffer.h"																								//	include the header

#define DEFAULT_CHAR_BUFFER_SIZE 2048									

////////////////////////////////////////////////////////////////////////
//									  																								    //
//	Member Functions:  																						       //
//									  																								    //
//		Constructors / Destructors / Disposal / Copying / Creation			           //
//									  																								    //
////////////////////////////////////////////////////////////////////////

CharBuffer::CharBuffer()																							//	default constructor
	{
	theData = new char[DEFAULT_CHAR_BUFFER_SIZE];												//	allocate suitable chunk of memory
	if (theData == NULL) throw (OutOfMemory());													//	if it fails, throw an exception
	readHead = theData;																									//	start the read head at the beginning
	writeHead = theData;																								//	and the write head
	*writeHead = '\0';																									//	be paranoid:  null-terminate it
	maxSize = DEFAULT_CHAR_BUFFER_SIZE;																	//	and set the size
	size = 1;																														//	initial size includes the null
	}

CharBuffer::CharBuffer(long newSize)																	//	constructor
	{
	theData = new char[newSize];																				//	allocate suitable chunk of memory
	if (theData == NULL) throw (OutOfMemory());													//	if it fails, throw an exception
	readHead = theData;																									//	start the read head at the beginning
	writeHead = theData;																								//	and the write head
	*writeHead = '\0';																									//	be paranoid:  null-terminate it
	maxSize = newSize;																									//	and set the size
	size = 1;																														//	initial size includes the null
	}

CharBuffer::CharBuffer(char *newContents)															//	constructor to take an existing string
	{
	theData = newContents;																							//	point to the string
	maxSize = 1 + strlen(newContents);																	//	set the maximum size
	size = maxSize;																											//	and set the existing size
	writeHead = theData + size - 1;																			//	set the write head
	readHead = theData;																									//	set the read head to the front
	}	

CharBuffer::~CharBuffer()																							//	destructor
	{
	delete theData;																											//	release the buffer
	}	

////////////////////////////////////////////////////////////////////////
//									  																								//
//	File Output:																											//
//									  																								//
////////////////////////////////////////////////////////////////////////
	
void CharBuffer::Write(char *newStuff)																//	adds more stuff at the entry point
	{
	char *newBuffer;																										//	for expanding, if necessary
	long newStuffLength = strlen(newStuff);															//	find out how long the new string is
	
	if (size + newStuffLength > maxSize)																//	if that is going to put us over our limit
		{
		for (; maxSize < size	+ newStuffLength; maxSize << 1){};					//	find a size that is big enough
		newBuffer = new char[maxSize];																		//	allocate a new buffer there
		if (newBuffer == NULL) throw (OutOfMemory());											//	if it comes back empty, throw a memory error
		strcpy(newBuffer, theData);																				//	copy the old stuff to the new buffer
		delete theData;																										//	dispose of the old copy
		writeHead = newBuffer + size - 1;																	//	entry point is at the null
		}
	strcpy(writeHead, newStuff);																				//	add the new stuff at the entry point			
	writeHead += newStuffLength;																				//	reset the entry point
	size += newStuffLength;																							//	and the size
	}		

char *CharBuffer::ReadLine()																					//	returns a copy of the data from the entry point to 
																																			//	the next new line or null character.  Null-terminates
																																			//	the string (destroys the newline character)
	{
	char *end;																													//	the tail end of the line to return
	char *returnString;																									//	the copy to return to the caller
	long lineLength;																										//	length of the line to copy
	
	end = strpbrk(readHead, "\n");																			//	find the next newline or null
	if (end == NULL)																										//	if there isn't one . . .
		end = writeHead;																									//	then set the end to the end of the string
	lineLength = end - readHead;																				//	number of characters that will need copying
	returnString = new char[lineLength+1];															//	allocate a new string for it
	strncpy(returnString, readHead, lineLength);												//	copy all of the characters to the new string
	returnString[lineLength] = '\0';																		//	and make sure that it is null-terminated
	if (*end == '\0')																										//	if we hit the end of the line
		readHead = end;																										//	set the entry point to it
	else
		readHead = end + 1;					  																		//	advance the entry point
	return returnString;																								//	and return the string	
	}	

bool CharBuffer::EndOfBuffer()																				//	returns true iff the entry point is at the end
	{
	return (readHead == writeHead);																			//	do the test & return the value
	}
	
long CharBuffer::GetLong()																						//	retrieve a long value
	{
	long returnValue;																										//	the value to return
	char *seek;																													//	pointer to the remainder of the buffer
	
	returnValue = strtol(readHead, &seek, 10);													//	read in the value
	if (readHead == seek)																								//	this happens if the scan failed to find a number
		throw EndOfBufferHit();																						//	so we throw an exception

	readHead = seek;																										//	advance the entry point
	for (;(*readHead == ' ') || (*readHead == '\t'); ) readHead++;			//	walk past any whitespace
	return returnValue;																									//	otherwise, return whatever
	}
	
float CharBuffer::GetFloat()																					//	retrieve a floating point value
	{
	float returnValue;																									//	the value to return
	char *seek;																													//	pointer to the remainder of the buffer
	
	returnValue = strtod(readHead, &seek);															//	read in the value
	if (readHead == seek)																								//	this happens if the scan failed to find a number
		throw EndOfBufferHit();																						//	so we throw an exception

	readHead = seek;																										//	advance the entry point
	for (;(*readHead == ' ') || (*readHead == '\t'); ) readHead++;			//	walk past any whitespace
	return returnValue;																									//	otherwise, return whatever
	}
	
char *CharBuffer::GetContents()																				//	strips out the contents & returns them
	{
	char *temp;																													//	temporary variable to hold the contents
	
	temp = theData;																											//	make the temp point to the data
	theData = new char[DEFAULT_CHAR_BUFFER_SIZE];												//	give it a new buffer
	if (theData == NULL) throw (OutOfMemory());													//	if it fails, throw an exception
	readHead = theData;																									//	start the read head at the beginning
	writeHead = theData;																								//	and the write head
	*writeHead = '\0';																									//	be paranoid:  null-terminate it
	maxSize = DEFAULT_CHAR_BUFFER_SIZE;																	//	and set the size
	size = 1;																														//	initial size includes the null

	return temp;																												//	and return the original contents
	}

char CharBuffer::Lookahead()																					//	returns the next character without advancing
	{
	return (*readHead);																									//	do the test & return the value
	}
	
