//	74.442	Advanced Design & Analysis of Algorithms
//	Spring 1998
//	Dr. Bill Kocay
//
//	Hamish Carr
//	Project:	Toroidal Surface Geometries
//
//

#include "Torus.h"
	
bool Torus::MakeLine(Tile *firstPoint, Tile *secondPoint)					//	makes a line, based on the selection
	{
	long theta1, theta2, phi1, phi2;										//	coordinates of the points
	long x, y;																//	theta, phi distance between the points
	long gcdxy;																//	gcd of x, y
	long xReps, yReps;														//	number of iterations necessary in x, y directions
	long totalReps;															//	total number of reps
	Line *whichLine;														//	pointer to the line being created
	Segment *whichSeg;														//	pointer to a given segment
	Segment *stopSeg;														//	sentinel for pointer walk
	Segment *nextSeg;														//	logically next segment (cyclic)
	Tile *whichTile;														//	pointer to the tile we want to add segment to
	float slope;															//	slope of the line, in tile units
	long nUnits;															//	how many units per tile for our slope calculation
	long xStep, yStep;														//	distance to step in y (x) to next x (y) grid line
	long xCross, yCross;													//	where next x (y) grid line is crossed
	float t;																//	parameter for distance along side of tile
	
	long i, j, k;															//	indices for our walk around the torus
	long segNum;															//	segment number in the segment list
	
	//	start off by checking to make sure the line doesn't already exist
	for (i = 0; i < nLines; i++)											//	walk down the array
		if (PointOnLine(firstPoint, theLines + i) && PointOnLine(secondPoint, theLines + i))
			return false;													//	line already exists:  don't add it

	//	I need to figure out what the coordinates of each point are
	
	GetCoords(firstPoint, theta1, phi1);									//	get the coords of the first point
	GetCoords(secondPoint, theta2, phi2);									//	and of the second point
	
	x = phi2 - phi1; y = theta2 - theta1;									//	compute distance in each direction
	
	//	next I need to set up the line as an array of segments
	
	if (nLines >= maxNLines)												//	if the array of lines is already full
		{
		char message[100];													//	buffer for the message
		
		sprintf(message, "I am sorry.  I am only allowed to have %d lines", maxNLines);
		DoAlert(message);													//	show the alert
		return false;														//	and leave the routine
		} // end of case where too many lines

	whichLine = theLines+nLines;											//	get a pointer to the line to construct
	nLines++;																//	increment the logical number of lines
	
	selectedLines[nLinesSelected] = whichLine;								//	hook it into the array of selected lines
	nLinesSelected++;														//	increment the number of selected lines
	whichLine->isSelected = true;											//	mark it as selected

	whichLine->theta = theta1; whichLine->phi = phi1;						//	mark the line as passing through here

	if (x == 0)																//	if we have a line around the ring
		{
		whichSeg = new Segment;												//	there is only one segment
		whichLine->theSegments = whichSeg;									//	make the line point to it
		whichLine->nSegments = 1;											//	all segments are the same
		whichSeg->inTL = 0.5;												//	set the entry point
		whichSeg->inBL = 0.5;
		whichSeg->inTR = 0;
		whichSeg->inBR = 0;
		whichSeg->outTL = 0;
		whichSeg->outBL = 0;
		whichSeg->outTR = 0.5;
		whichSeg->outBR = 0.5;
		whichSeg->thetaIncrement = 1;										//	set increments for walk around torus
		whichSeg->phiIncrement = 0;
		totalReps = nThetaDivisions;										//	how many repetitions of the basic line
		whichLine->dTheta = 1;												//	set the pitch in the line structure
		whichLine->dPhi = 0;		
		} // end of case where we have a line around the ring
	else if (y == 0)														//	i.e. we have a line around the tube	
		{
		whichSeg = new Segment;												//	there is only one segment
		whichLine->theSegments = whichSeg;									//	make the line point to it
		whichLine->nSegments = 1;											//	all segments are the same
		whichSeg->inTL = 0.5;												//	set the entry point
		whichSeg->inTR = 0.5;
		whichSeg->inBL = 0;
		whichSeg->inBR = 0;
		whichSeg->outTL = 0;
		whichSeg->outTR = 0;
		whichSeg->outBL = 0.5;
		whichSeg->outBR = 0.5;
		whichSeg->thetaIncrement = 0;										//	set increments for walk around torus
		whichSeg->phiIncrement = 1;
		totalReps = nPhiDivisions;											//	how many repetitions of the basic line
		whichLine->dTheta = 0;												//	set the pitch in the line structure
		whichLine->dPhi = 1;		
		} // end of case where we have a line around the tube	
	else																	//	i.e. we have a spiral	
		{
		if (2*x > nPhiDivisions)											//	if it's more than halfway
			x = x - nPhiDivisions;											//	loop the other way
		
		if (2*y > nThetaDivisions)											//	if it's more than halfway
			y = y - nThetaDivisions;										//	loop the other way
	
		gcdxy = GCD(abs(x), abs(y));										//	find the GCD
	
		if (gcdxy != 1)														//	if they are not relatively prime
			{
			x /= gcdxy; y /= gcdxy;											//	scale them so they are relatively prime
			} // end of test for gcdxy != 1
		
		whichLine->dTheta = y;												//	set the pitch in the line structure
		whichLine->dPhi = x;		

		xReps = nPhiDivisions / GCD(nPhiDivisions, abs(x));					//	calculate # of loops around tube to line up neatly	
		yReps = nThetaDivisions / GCD(nThetaDivisions, abs(y));				//	and the # of times around the ring to line up neatly
	
		totalReps = LCM(xReps, yReps);										//	calculate total # of repetitions of basic distance

		if ((abs(x)%2 == 1) && (abs(y)%2 == 1))								//	if both are odd, it passes exactly through a vertex
			{
			whichLine->nSegments = abs(x) + abs(y) - 1;						//	-1 because we pass through a vertex
			whichLine->theSegments = new Segment[whichLine->nSegments];		//	allocate space for them
			} // end of case where both odd
		else																//	at least one is even
			{
			whichLine->nSegments = abs(x) + abs(y);							//	we need exactly x + y segments
			whichLine->theSegments = new Segment[whichLine->nSegments];		//	allocate space for them
			} // end of case where either is even
		
		totalReps *= whichLine->nSegments;									//	combine the two factors

		nUnits = abs(2 * x * y);											//	# of fractions/tile - common denominator
		xStep = 2 * y * y;													//	actually n (x/y) simplified
		yStep = 2 * x * x;													//	n (y/x) simplified

		xCross = (nUnits + xStep)/2;										//	find the first x gridline crossed
		yCross = (nUnits + yStep)/2;										//	and the first y gridline crossed
		
		slope = y * (1.0 / x);												//	calculate the slope - 0's taken care of above
		
		stopSeg = whichLine->theSegments + whichLine->nSegments;			//	calculate sentinel
		for (whichSeg = whichLine->theSegments; whichSeg < stopSeg; whichSeg++)
																			//	loop until all segments constructed
			{
			nextSeg = whichSeg + 1;											//	step to next segment
			if (nextSeg == stopSeg)											//	if it's the sentinel
				nextSeg = whichLine->theSegments;							//	use the first one instead

			if (xCross < nUnits)											//	i.e. if x grid is crossed first
				{
				t = xCross*(1.0/nUnits);									//	parameterize the position
				if (slope > 0)
					{
					whichSeg->outTL = whichSeg->outTR = 0;					//	we don't use TL or TR
					whichSeg->outBR = t;									//	and xCross represents the BL component
					whichSeg->outBL = 1 - t;								//	the rest belongs to BR
					whichSeg->phiIncrement = 1;								//	segment does change phis
					nextSeg->inTR = whichSeg->outBR;						//	set the next segments inputs
					nextSeg->inTL = whichSeg->outBL;
					nextSeg->inBR = nextSeg->inBL = 0;						//	unused components
					} // end of positive slope					
				else														//	i.e. negative slope
					{
					whichSeg->outBL = whichSeg->outBR = 0;					//	we don't use BL or BR
					whichSeg->outTR = t;									//	and xCross represents the TL component
					whichSeg->outTL = 1 - t;								//	the rest belongs to TR
					whichSeg->phiIncrement = -1;							//	segment does change phis
					nextSeg->inBR = whichSeg->outTR;						//	set the next segments inputs
					nextSeg->inBL = whichSeg->outTL;
					nextSeg->inTR = nextSeg->inTL = 0;						//	unused components
					} // end of negative slope
				xCross += xStep;											//	find the next x crossing
				yCross -= nUnits;											//	and move the y crossing over
				whichSeg->thetaIncrement = 0;								//	the segment doesn't change thetas
				} // end of x crossed first
			else if (yCross < nUnits)										//	i.e. if y grid is crossed first
				{
				t = yCross * (1.0/nUnits);									//	parameterize the position
				if (slope > 0)												//	i.e. positive slope
					{
					whichSeg->outBR = t;									//	and yCross represents the BR component
					whichSeg->outTR = 1 - t;								//	the rest belongs to TR
					} // end of positive slope					
				else														//	i.e. negative slope
					{
					whichSeg->outTR = t;									//	and yCross represents the TR component
					whichSeg->outBR = 1 - t;								//	the rest belongs to BR
					} // end of negative slope
				whichSeg->outTL = whichSeg->outBL = 0;						//	we don't use TL or BL
				yCross += yStep;											//	find the next y crossing
				xCross -= nUnits;											//	and move the y crossing over
				whichSeg->thetaIncrement = 1;								//	the segment changes thetas
				whichSeg->phiIncrement = 0;									//	and doesn't change phis
				nextSeg->inTL = whichSeg->outTR;							//	and set the next segment's inputs
				nextSeg->inBL = whichSeg->outBR;		
				nextSeg->inTR = nextSeg->inBR = 0;							//	we don't use these						
				} // end of y crossed first
			else															//	i.e. both cross exactly at corner
				{
				if (slope > 0)												//	i.e. positive slope
					{
					whichSeg->outTR = 0; whichSeg->outBR = 1; 				//	the rest are determined easily
					whichSeg->phiIncrement = 1;								//	segment changes phis
					} // end of positive slope					
				else														//	i.e. negative slope
					{
					whichSeg->outBR = 0; whichSeg->outTR = 1; 				//	the rest are determined easily
					whichSeg->phiIncrement = -1;							//	segment changes phis
					} // end of negative slope
				whichSeg->outTL = 0; whichSeg->outBL = 0;					//	we don't use TL or BL
				xCross = xStep;												//	find the next x crossing
				yCross = yStep;												//	find the next y crossing
				whichSeg->thetaIncrement = 1;								//	set the theta step
				nextSeg->inTL = whichSeg->outBR;							//	set the next segment's inputs
				nextSeg->inBL = whichSeg->outTR;
				nextSeg->inTR = nextSeg->inBR = 0;							//	unused components
				} // end of case where both cross simultaneously
			} // end of loop to construct segments
		} // end of case where line is a spiral
	
	// next we need to add the line to the tiles
	// essentially, we walk around, adding it to one tile at a time
	// until we return to the beginning
	
	i = theta1; j = phi1;													//	start at the first point
	segNum = 0;																//	and the first (i.e. 0th) segment
	
	for (k = 0; k < totalReps; k++)											//	repeat this loop
		{
		whichSeg = whichLine->theSegments + segNum;							//	retrieve the corresponding segment
		whichTile = GetTile(i, j);											//	retrieve the correct tile (could be pointer walk)
		whichTile->AddSegment(whichSeg);									//	add the segment to the tile
		
		i += whichSeg->thetaIncrement;										//	use the segment to determine direction of step
		j += whichSeg->phiIncrement;										//	in both directions
		segNum++;															//	always move to the next segment
		
		if (i >= nThetaDivisions) i -= nThetaDivisions;						//	chop it down to size
		if (j >= nPhiDivisions) j -= nPhiDivisions;							//	chop this one too
		if (i < 0) i += nThetaDivisions;									//	chop it up to size
		if (j < 0) j += nPhiDivisions;										//	chop this one too
		if (segNum >= whichLine->nSegments) segNum -= whichLine->nSegments;	//	and this one
		} // end of for loop
	return true;															//	and return a happy camper code
	} // end of MakeLine()
	
bool Torus::MakeLineFromSelection()											//	makes lines, based on the selection of points
	{
	if (nPointsSelected < 2)												//	i.e. we don't have exactly two nodes selected
		{
		DoAlert("At least 2 points must be selected");						//	pop up an alert box
		return false;														//	and advise the view not to update
		} // end of case where less than two points selected
	else
		for (long i = 0; i < nPointsSelected; i++)							//	loop through all points
			for (long j = i+1; j < nPointsSelected; j++)					//	loop through all higher points
				MakeLine(selectedPoints[i], selectedPoints[j]);				//	make a line through the pair
	return true;															//	return true
	} // end of MakeLineFromSelection()

bool Torus::DeleteSelectedLines() 											//	deletes all selected lines
	{
	return true;
	} // end of DeleteSelectedLines()

bool Torus::DeleteLine(long whichLine)										//	deletes the specified line
	{
	#pragma unused(whichLine)
	return true;
	} // end of DeleteLine()

bool Torus::PointOnLine(Tile *thePoint, Line *theLine)						//	tests whether a point is on a line
	{
	SegmentNode *walker;													//	for walking down the list of segments

	for (walker = thePoint->segmentList; walker != NULL; walker = walker->nextSegment)
		{
		if (walker->theSegment == theLine->theSegments)						//	if it points to the first segment
			return true;													//	say that it is on the line
		} // end of walk down list of segments

	return false;															//	say that it is not on the line
	} // end of PointOnLine()
