//	74.442	Advanced Design & Analysis of Algorithms
//	Spring 1998
//	Dr. Bill Kocay
//
//	Hamish Carr
//	Project:	Toroidal Surface Geometries
//
//

#include "TorusView.h"

TorusView::TorusView(BRect theFrame, Torus *newTorus)						//	constructor
	: BView(theFrame, "Torus Main View", B_FOLLOW_ALL, B_WILL_DRAW)			//	call inherited constructor
	{
	theTorus = newTorus;													//	set the torus to the one passed in
	BRect theRect = theFrame;												//	set a rectangle to the frame
	theRect.OffsetTo(0.0,0.0);												//	make the origin the top right

	theBlitMap = new BBitmap(theRect, B_RGB_32_BIT, true);					//	create blitmap
	theBlitView = new TorusBlitView(theRect, theTorus, theBlitMap);			//	create the blitview
	theBlitMap->AddChild(theBlitView);										//	add the blitview to the blitmap
	theBlitView->Generate();												//	fill in the blitmap
	} // end of TorusView constructor

TorusView::~TorusView()														//	destructor
	{
	theBlitMap->Lock();														//	lock the blitmap first
	delete theBlitMap;
	} // end of TorusView destructor

void TorusView::Draw(BRect updateRect)										//	draw the torus
	{
	DrawBitmap(theBlitMap, updateRect, updateRect);							//	blit the bits
	} // end of Draw()

void TorusView::MouseDown(BPoint where)										//	handle a mouse click
	{
	bool viewChanged = false;												//	used for capturing whether to change view
	bool shiftDown;															//	whether the shift key is down
	long modifiers;															//	used to test which buttons are down

	Window()->CurrentMessage()->FindInt32("modifiers", &modifiers);			//	retrieve the current button state
	shiftDown = modifiers & B_SHIFT_KEY;									//	work out whether shift key is down
	switch (theTorus->theCurrentTool)										//	depending on which tool is selected
		{
		case selectionTool:													//	selects a point
			viewChanged = theTorus->TogglePointSelection(where, shiftDown);	//	toggle the selection there
			break;
		case pointTool:														//	creates a new point
			viewChanged = theTorus->TogglePoint(where, shiftDown);			//	toggle the existence of the point
			break;
		case handTool:														//	moves the entire object
			break;
		} // end of cases for various tools
	if (viewChanged)														//	check to see if we need to update view
		{
		theBlitView->Generate();											//	generate a new blit map
		theTorusToolView->Draw(theTorusToolView->Bounds());					//	redraw the tool palette
		Draw(Bounds());														//	and the main view
		} // end of case where view changes
	} // end of MouseDown()

void TorusView::KeyDown(const char *bytes, int32 numBytes)					//	handle a keystroke
	{
	Basis rotationBasis;													//	the rotation to perform
	
	if (numBytes == 1)														//	if it's a single character
		{
		switch(bytes[0])													//	cases for each one that interests us
			{
			case B_UP_ARROW:											
				theTorus->RotateTiles(RotateUpItem);						//	rotate the tiles up
				break;
			case B_DOWN_ARROW:												
				theTorus->RotateTiles(RotateDownItem);						//	tell the window to rotate the torus tiles down
				break;
			case B_LEFT_ARROW:												
				theTorus->RotateTiles(RotateLeftItem);						//	tell the window to rotate the torus tiles left
				break;
			case B_RIGHT_ARROW:										
				theTorus->RotateTiles(RotateRightItem);						//	tell the window to rotate the torus tiles right
				break;
			case '2':
				rotationBasis = Basis(Vector(1, 0, 0), Vector(0, cos(0.05), sin(0.05)), Vector(0, -sin(0.05), cos(0.05)));
				theTorus->Rotate(rotationBasis);
				break;
			case '4':
				rotationBasis = Basis(Vector(cos(0.05), -sin(0.05), 0), Vector(sin(0.05), cos(0.05), 0), Vector(0, 0, 1));
				theTorus->Rotate(rotationBasis);
				break;
			case '5':
				theTorus->TranslateZ(100);
				break;
			case '6':
				rotationBasis = Basis(Vector(cos(0.05), sin(0.05), 0), Vector(-sin(0.05), cos(0.05), 0), Vector(0, 0, 1));
				theTorus->Rotate(rotationBasis);
				break;
			case '7':
				rotationBasis = Basis(Vector(cos(0.05), 0, sin(0.05)), Vector(0, 1, 0), Vector(-sin(0.05), 0, cos(0.05)));
				theTorus->Rotate(rotationBasis);
				break;
			case '8':
				rotationBasis = Basis(Vector(1, 0, 0), Vector(0, cos(0.05), -sin(0.05)), Vector(0, sin(0.05), cos(0.05)));
				theTorus->Rotate(rotationBasis);
				break;
			case '9':
				rotationBasis = Basis(Vector(cos(0.05), 0, -sin(0.05)), Vector(0, 1, 0), Vector(sin(0.05), 0, cos(0.05)));
				theTorus->Rotate(rotationBasis);
				break;
			case '0':
				theTorus->TranslateZ(-100);
				break;
			default:														//	anything we don't want
				return;														//	leave without redrawing
			} // end of switch statement		
			theBlitView->Generate();										//	redraw the views
			Draw(Bounds());		
			theTorusToolView->Draw(theTorusToolView->Bounds());	
		} // end of case for single byte character
	} // end of KeyDown()
	
void TorusView::MessageReceived(BMessage *theEvent)							//	processes messages
	{
	bool viewChanged = false;												//	used for capturing whether to change view
	switch (theEvent->what)													//	depending on what the message is
		{	
		case ConnectItem:													//	if the user asked us to connect lines
			viewChanged = theTorus->MakeLineFromSelection();				//	do so
			break;
		case RotateUpItem:													//	rotates outside face up round tube
		case RotateDownItem:												//	rotates outside face down round tube
		case RotateLeftItem:												//	rotates points left round torus
		case RotateRightItem:												//	rotates points right round torus
			viewChanged = theTorus->RotateTiles(theEvent->what);			//	pass the request on as a message
			break;
		default:															//	all other cases
			BView::MessageReceived(theEvent);								//	call the inherited routine
			break;
		} // end of cases of messages
	if (viewChanged)														//	check to see if we need to update view
		{
		theBlitView->Generate();											//	generate a new blit map
		theTorusToolView->Draw(theTorusToolView->Bounds());					//	redraw the tool palette
		Draw(Bounds());														//	and the main view
		} // end of case where view changes
	} // end of MessageReceived()